import logging
import argparse
import sys
import time
from collections import defaultdict, namedtuple
import urlparse
from copy import copy

from gevent import monkey
import gevent
from gevent.pool import Pool
from gevent.socket import gethostbyname

monkey.patch_all()

import requests
from requests import RequestException

from boom import __version__, _patch     # NOQA
from boom.util import resolve_name
from boom.pgbar import AnimatedProgressBar

try:
    from gevent.dns import DNSError
except ImportError:
    from socket import error as DNSError

logger = logging.getLogger('boom')

_VERBS = ('GET', 'POST', 'DELETE', 'PUT', 'HEAD', 'OPTIONS')
_DATA_VERBS = ('POST', 'PUT')


class RunResults(object):
    """Encapsulates the results of a single Boom run.

    Contains a dictionary of status codes to lists of request durations,
    a list of exception instances raised during the run, the total time
    of the run and an animated progress bar.
    """

    def __init__(self, num=1, quiet=False):
        self.status_code_counter = defaultdict(list)
        self.errors = []
        self.total_time = None
        if num is not None:
            self._progress_bar = AnimatedProgressBar(end=num, width=65)
        else:
            self._progress_bar = None
        self.quiet = quiet

    def incr(self):
        if self.quiet:
            return
        if self._progress_bar is not None:
            self._progress_bar + 1
            self._progress_bar.show_progress()
        else:
            sys.stdout.write('.')
            sys.stdout.flush()


RunStats = namedtuple('RunStats', ['count', 'total_time', 'rps', 'avg', 'min',
                                   'max', 'amp'])


def calc_stats(results):
    """Calculate stats (min, max, avg) from the given RunResults.

       The statistics are returned as a RunStats object.
    """
    all_res = []
    count = 0
    for values in results.status_code_counter.values():
        all_res += values
        count += len(values)

    cum_time = sum(all_res)

    if cum_time == 0 or len(all_res) == 0:
        rps = avg = min_ = max_ = amp = 0
    else:
        if results.total_time == 0:
            rps = 0
        else:
            rps = len(all_res) / float(results.total_time)
        avg = sum(all_res) / len(all_res)
        max_ = max(all_res)
        min_ = min(all_res)
        amp = max(all_res) - min(all_res)
    return RunStats(count, results.total_time, rps, avg, min_, max_, amp)


def print_stats(results):
    stats = calc_stats(results)
    rps = stats.rps

    print('')
    print('-------- Results --------')

    print('Successful calls\t\t%r' % stats.count)
    print('Total time       \t\t%.4f s' % stats.total_time)
    print('Average          \t\t%.4f s' % stats.avg)
    print('Fastest          \t\t%.4f s' % stats.min)
    print('Slowest          \t\t%.4f s' % stats.max)
    print('Amplitude        \t\t%.4f s' % stats.amp)
    print('RPS              \t\t%d' % rps)
    if rps > 500:
        print('BSI              \t\tWoooooo Fast')
    elif rps > 100:
        print('BSI              \t\tPretty good')
    elif rps > 50:
        print('BSI              \t\tMeh')
    else:
        print('BSI              \t\tHahahaha')
    print('')
    print('-------- Status codes --------')
    for code, items in results.status_code_counter.items():
        print('Code %d          \t\t%d times.' % (code, len(items)))
    print('')
    print('-------- Legend --------')
    print('RPS: Request Per Second')
    print('BSI: Boom Speed Index')


def print_server_info(url, method, headers=None):
    res = requests.head(url)
    print 'Server Software: %s' % res.headers.get('server', 'Unknown')

    print 'Running %s %s' % (method, url)

    if headers:
        for k, v in headers.items():
            print '\t%s: %s' % (k, v)


def print_errors(errors):
    if len(errors) == 0:
        return
    print('')
    print('-------- Errors --------')
    for error in errors:
        print(error)


def print_json(results):
    """Prints a JSON representation of the results to stdout."""
    import json
    stats = calc_stats(results)
    print(json.dumps(stats._asdict()))


def onecall(method, url, results, **options):
    """Performs a single HTTP call and puts the result into the
       status_code_counter.

    RequestExceptions are caught and put into the errors set.
    """
    start = time.time()

    if 'data' in options and callable(options['data']):
        options = copy(options)
        options['data'] = options['data'](method, url, options)

    if 'hook' in options:
        method, url, options = options['hook'](method, url, options)
        del options['hook']

    try:
        res = method(url, **options)
    except RequestException as exc:
        results.errors.append(exc)
    else:
        duration = time.time() - start
        results.status_code_counter[res.status_code].append(duration)
    finally:
        results.incr()


def run(url, num=1, duration=None, method='GET', data=None, ct='text/plain',
        auth=None, concurrency=1, headers=None, hook=None, quiet=False):

    if headers is None:
        headers = {}

    if 'content-type' not in headers:
        headers['Content-Type'] = ct

    if data is not None and data.startswith('py:'):
        callable = data[len('py:'):]
        data = resolve_name(callable)

    method = getattr(requests, method.lower())
    options = {'headers': headers}

    if hook is not None:
        options['hook'] = resolve_name(hook)

    if data is not None:
        options['data'] = data

    if auth is not None:
        options['auth'] = tuple(auth.split(':', 1))

    pool = Pool(concurrency)

    start = time.time()
    jobs = None

    res = RunResults(num, quiet)

    try:
        if num is not None:
            jobs = [pool.spawn(onecall, method, url, res, **options)
                    for i in range(num)]
            pool.join()
        else:
            with gevent.Timeout(duration, False):
                jobs = []
                while True:
                    jobs.append(pool.spawn(onecall, method, url, res,
                                           **options))
                pool.join()
    except KeyboardInterrupt:
        # In case of a keyboard interrupt, just return whatever already got
        # put into the result object.
        pass
    finally:
        res.total_time = time.time() - start

    return res


def resolve(url):
    parts = urlparse.urlparse(url)
    netloc = parts.netloc.rsplit(':')

    if len(netloc) == 1:
        if parts.scheme == 'https':
            netloc.append('443')
        else:
            netloc.append('80')

    original = netloc[0]
    resolved = gethostbyname(original)

    # Don't use a resolved hostname for SSL requests otherwise the
    # certificate will not match the IP address (resolved)
    if parts.scheme == 'https':
        resolved = original

    netloc = resolved + ':' + netloc[1]
    parts = (parts.scheme, netloc) + parts[2:]
    return urlparse.urlunparse(parts), original, resolved


def load(url, requests, concurrency, duration, method, data, ct, auth,
         headers=None, hook=None, quiet=False):
    if not quiet:
        print_server_info(url, method, headers=headers)

        if requests is not None:
            print('Running %d times per %d workers.' % (requests, concurrency))
        else:
            print('Running %d workers for at least %d seconds.' %
                  (concurrency, duration))

        sys.stdout.write('Starting the load')
    try:
        return run(url, requests, duration, method, data, ct,
                   auth, concurrency, headers, hook, quiet=quiet)
    finally:
        if not quiet:
            print(' Done')


def main():
    parser = argparse.ArgumentParser(description='Simple HTTP Load runner.')

    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays version and exits.')

    parser.add_argument('-m', '--method', help='HTTP Method',
                        type=str, default='GET', choices=_VERBS)

    parser.add_argument('--content-type', help='Content-Type',
                        type=str, default='text/plain')

    parser.add_argument('-D', '--data',
                        help=('Data. Prefixed by "py:" to point '
                              'a python callable.'),
                        type=str)

    parser.add_argument('-c', '--concurrency', help='Concurrency',
                        type=int, default=1)

    parser.add_argument('-a', '--auth',
                        help='Basic authentication user:password', type=str)

    parser.add_argument('--header', help='Custom header. name:value',
                        type=str, action='append')

    parser.add_argument('--hook',
                        help=("Python callable that'll be used "
                              "on every requests call"),
                        type=str)

    parser.add_argument('--json-output',
                        help='Prints the results in JSON instead of the '
                             'default format',
                        action='store_true')

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-n', '--requests', help='Number of requests',
                       type=int)

    group.add_argument('-d', '--duration', help='Duration in seconds',
                       type=int)

    parser.add_argument('url', help='URL to hit', nargs='?')
    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    if args.url is None:
        print('You need to provide an URL.')
        parser.print_usage()
        sys.exit(0)

    if args.data is not None and not args.method in _DATA_VERBS:
        print("You can't provide data with %r" % args.method)
        parser.print_usage()
        sys.exit(0)

    if args.requests is None and args.duration is None:
        args.requests = 1

    try:
        url, original, resolved = resolve(args.url)
    except DNSError, e:
        print_errors(("DNS resolution failed for %s (%s)" %
                      (args.url, str(e)),))
        sys.exit(1)

    def _split(header):
        header = header.split(':')

        if len(header) != 2:
            print("A header must be of the form name:value")
            parser.print_usage()
            sys.exit(0)

        return header

    if args.header is None:
        headers = {}
    else:
        headers = dict([_split(header) for header in args.header])

    if original != resolved and 'Host' not in headers:
        headers['Host'] = original

    try:
        res = load(url, args.requests, args.concurrency, args.duration,
                   args.method, args.data, args.content_type, args.auth,
                   headers=headers, hook=args.hook, quiet=args.json_output)
    except RequestException as e:
        print_errors((e, ))
        sys.exit(1)

    if not args.json_output:
        print_errors(res.errors)
        print_stats(res)
    else:
        print_json(res)
    logger.info('Bye!')


if __name__ == '__main__':
    main()



import unittest2 as unittest
import subprocess
import sys
import shlex
import StringIO
import json

from gevent.pywsgi import WSGIServer
import requests
import gevent

from boom._boom import run as runboom, main, resolve, RunResults
from boom import _boom


class App(object):
    def __init__(self):
        self.numcalls = 0

    def handle(self, env, start_response):
        if env['PATH_INFO'] == '/':
            self.numcalls += 1
            start_response('200 OK', [('Content-Type', 'text/html')])
            return ["<b>hello world</b>"]
        elif env['PATH_INFO'] == '/calls':
            start_response('200 OK', [('Content-Type', 'text/plain')])
            return [str(self.numcalls)]
        elif env['PATH_INFO'] == '/redir':
            self.numcalls += 1
            start_response('302 Found', [('Location', '/redir')])
            return []
        elif env['PATH_INFO'] == '/reset':
            self.numcalls = 0
            start_response('200 OK', [('Content-Type', 'text/plain')])
            return ['numcalls set to zero']
        else:
            start_response('404 Not Found', [('Content-Type', 'text/plain')])
            return ['%s' % env['PATH_INFO']]


def run():
    app = App()
    WSGIServer(('0.0.0.0', 8089), app.handle).serve_forever()


_CMD = "%s -c 'from boom.tests.test_boom import run; run()'"
CMD = shlex.split(_CMD % sys.executable)
_SERVER = None


def _start():
    global _SERVER
    _SERVER = subprocess.Popen(CMD, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)


def _stop():
    global _SERVER
    if _SERVER is not None:
        _SERVER.terminate()
        _SERVER = None


def hook(method, url, options):
    options['files'] = {'file': open(__file__, 'rb')}
    return method, url, options


class TestBoom(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        _start()
        cls.server = 'http://0.0.0.0:8089'
        while True:
            try:
                requests.get(cls.server + '/')
                return
            except requests.ConnectionError:
                gevent.sleep(.1)

    @classmethod
    def tearDownClass(cls):
        _stop()

    def setUp(self):
        self.get('/reset')

    def get(self, path):
        return requests.get(self.server + path)

    def test_basic_run(self):
        runboom(self.server, num=10, concurrency=1, quiet=True)
        res = self.get('/calls').content
        self.assertEqual(int(res), 10)

    def test_hook(self):
        runboom(self.server, method='POST', num=10, concurrency=1,
                hook='boom.tests.test_boom.hook', quiet=True)
        res = self.get('/calls').content
        self.assertEqual(int(res), 10)

    def test_connection_error(self):
        run_results = runboom('http://localhost:9999', num=10, concurrency=1,
                              quiet=True)
        self.assertEqual(len(run_results.errors), 10)
        for error in run_results.errors:
            self.assertIsInstance(error, requests.ConnectionError)

    def test_too_many_redirects(self):
        run_results = runboom(self.server + '/redir', num=2, concurrency=1,
                              quiet=True)
        res = self.get('/calls').content
        self.assertEqual(int(res), 62)
        for error in run_results.errors:
            self.assertIsInstance(error, requests.TooManyRedirects)

    def _run(self, *args):
        sys.argv[:] = [sys.executable] + list(args)
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        exit_code = 0
        sys.stdout = StringIO.StringIO()
        sys.stderr = StringIO.StringIO()
        try:
            main()
        except (Exception, SystemExit), e:
            if isinstance(e, SystemExit):
                exit_code = e.code
            else:
                exit_code = 1
        finally:
            sys.stdout.seek(0)
            stdout = sys.stdout.read()
            sys.stderr.seek(0)
            stderr = sys.stdout.read()
            sys.stdout = old_stdout
            sys.stderr = old_stderr

        return exit_code, stdout, stderr

    def test_dns_resolve(self):
        code, stdout, stderr = self._run('http://that.impossiblename')
        self.assertEqual(code, 1)
        self.assertTrue('DNS resolution failed for' in stdout, stdout)

    def test_resolve(self):
        test_url = 'http://localhost:9999'
        url, original, resolved = resolve(test_url)
        self.assertEqual(url, 'http://127.0.0.1:9999')
        self.assertEqual(original, 'localhost')
        self.assertEqual(resolved, '127.0.0.1')

    def test_ssl_resolve(self):
        test_url = 'https://localhost:9999'
        url, original, resolved = resolve(test_url)
        self.assertEqual(url, 'https://localhost:9999')
        self.assertEqual(original, 'localhost')
        self.assertEqual(resolved, 'localhost')

    def test_resolve_no_scheme(self):
        test_url = 'http://localhost'
        url, original, resolved = resolve(test_url)
        self.assertEqual(url, 'http://127.0.0.1:80')
        self.assertEqual(original, 'localhost')
        self.assertEqual(resolved, '127.0.0.1')

    def test_resolve_no_scheme_ssl(self):
        test_url = 'https://localhost'
        url, original, resolved = resolve(test_url)
        self.assertEqual(url, 'https://localhost:443')
        self.assertEqual(original, 'localhost')
        self.assertEqual(resolved, 'localhost')

    def test_json_output(self):
        results = RunResults()
        results.status_code_counter['200'].extend([0, 0.1, 0.2])
        results.total_time = 9

        old_stdout = sys.stdout
        old_stderr = sys.stderr

        try:
            sys.stdout = StringIO.StringIO()
            sys.stderr = StringIO.StringIO()
            _boom.print_json(results)

            sys.stdout.seek(0)
            output = sys.stdout.read()
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr

        actual = json.loads(output)
        self.assertEqual(3, actual['count'])
        self.assertAlmostEqual(0.333, actual['rps'], delta=0.1)
        self.assertEqual(0, actual['min'])
        self.assertEqual(0.2, actual['max'])
        self.assertAlmostEqual(0.1, actual['avg'], delta=0.1)
        self.assertEqual(0.2, actual['amp'])


if __name__ == '__main__':
    run()



import threading
from threading import (_active_limbo_lock, _limbo, _active, _sys, _trace_hook,
                       _profile_hook, _format_exc)


# see http://bugs.python.org/issue1596321
def _bootstrap_inner(self):
    try:
        self._set_ident()
        self._Thread__started.set()
        with _active_limbo_lock:
            _active[self._Thread__ident] = self
            del _limbo[self]

        if _trace_hook:
            _sys.settrace(_trace_hook)
        if _profile_hook:
            _sys.setprofile(_profile_hook)

        try:
            self.run()
        except SystemExit:
            pass
        except:
            if _sys:
                _sys.stderr.write("Exception in thread %s:\n%s\n" %
                                  (self.name, _format_exc()))
            else:
                exc_type, exc_value, exc_tb = self._exc_info()
                try:
                    self._stderr.write(
                        "Exception in thread " + self.name + " (most likely "
                        "raised during interpreter shutdown):")

                    self._stderr.write("Traceback (most recent call last):")
                    while exc_tb:
                        self._stderr.write(
                            '  File "%s", line %s, in %s' %
                            (exc_tb.tb_frame.f_code.co_filename,
                                exc_tb.tb_lineno,
                                exc_tb.tb_frame.f_code.co_name))

                        exc_tb = exc_tb.tb_next
                    self._stderr.write("%s: %s" % (exc_type, exc_value))
                finally:
                    del exc_type, exc_value, exc_tb
        finally:
            pass
    finally:
        with _active_limbo_lock:
            self._Thread__stop()
            try:
                del _active[self._Thread__ident]
            except:
                pass


def _delete(self):
    try:
        with _active_limbo_lock:
            del _active[self._Thread__ident]
    except KeyError:
        if 'dummy_threading' not in _sys.modules:
            raise


# http://bugs.python.org/issue14308
def _stop(self):
    # DummyThreads delete self.__block, but they have no waiters to
    # notify anyway (join() is forbidden on them).
    if not hasattr(self, '_Thread__block'):
        return
    self._Thread__stop_old()


threading.Thread._Thread__bootstrap_inner = _bootstrap_inner
threading.Thread._Thread__delete = _delete
threading.Thread._Thread__stop_old = threading.Thread._Thread__stop
threading.Thread._Thread__stop = _stop



import unittest
from boom.pgbar import ProgressBar


class DefaultsTestCase(unittest.TestCase):
    """
    ProgressBar defaults:
    start = 0
    end = 10
    width = 12
    fill = '='
    blank = '.'
    format = '[%(fill)s>%(blank)s] %(progress)s%%'
    incremental = True
    """

    def setUp(self):
        self.p = ProgressBar()

    def tearDown(self):
        del (self.p)

    def test_initialization(self):
        """
        >>> p = ProgressBar()
        >>> p
        [>............] 0%
        """
        self.assertEqual(str(self.p), '[>............] 0%')

    def test_increment(self):
        """
        >>> p = ProgressBar()
        >>> p + 1
        [=>...........] 10%
        """
        self.p + 1
        self.assertEqual(str(self.p), '[=>...........] 10%')

    def test_reset(self):
        """
        >>> p = ProgressBar()
        >>> p += 8
        >>> p.reset()
        [>............] 0%
        """
        self.p += 8
        self.p.reset()
        self.assertEqual(str(self.p), '[>............] 0%')

    def test_full_progress(self):
        """
        >>> p = ProgressBar()
        >>> p + 10
        [============>] 100%
        """
        self.p + 10
        self.assertEqual(str(self.p), '[============>] 100%')
        self.p + 10
        self.assertEqual(str(self.p), '[============>] 100%')


class CustomizedTestCase(unittest.TestCase):
    """
    ProgressBar custom:
    start = 0
    end = 100
    width = 20
    fill = '#'
    blank = '.'
    format = '%(progress)s%% [%(fill)s%(blank)s]'
    incremental = True
    """
    custom = {
        'end': 100,
        'width': 20,
        'fill': '#',
        'format': '%(progress)s%% [%(fill)s%(blank)s]'
    }

    def setUp(self):
        self.p = ProgressBar(**self.custom)

    def tearDown(self):
        del (self.p)

    def test_initialization(self):
        """
        >>> custom = {
        ...  'end': 100,
        ...  'width': 20,
        ...  'fill': '#',
        ...  'format': '%(progress)s%% [%(fill)s%(blank)s]'
        ... }
        >>> p = ProgressBar(custom)
        >>> p
        0% [....................]
        """
        self.assertEqual(str(self.p), '0% [....................]')

    def test_increment(self):
        """
        >>> custom = {
        ...  'end': 100,
        ...  'width': 20,
        ...  'fill': '#',
        ...  'format': '%(progress)s%% [%(fill)s%(blank)s]'
        ... }
        >>> p = ProgressBar(custom)
        >>> p + 1
        1% [....................]
        """
        self.p + 1
        self.assertEqual(str(self.p), '1% [....................]')
        self.p + 4
        self.assertEqual(str(self.p), '5% [#...................]')

    def test_reset(self):
        """
        >>> custom = {
        ...  'end': 100,
        ...  'width': 20,
        ...  'fill': '#',
        ...  'format': '%(progress)s%% [%(fill)s%(blank)s]'
        ... }
        >>> p = ProgressBar(custom)
        >>> p += 8
        >>> p.reset()
        0% [....................]
        """
        self.p += 8
        self.p.reset()
        self.assertEqual(str(self.p), '0% [....................]')

    def test_full_progress(self):
        """
        >>> p = ProgressBar()
        >>> p + 10
        100% [####################]
        """
        self.p + 100
        self.assertEqual(str(self.p), '100% [####################]')
        self.p + 100
        self.assertEqual(str(self.p), '100% [####################]')


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
from boom import __version__
import sys

install_requires = ['gevent', 'requests', 'unittest2']

if sys.version_info < (2, 7):
    install_requires += ['argparse']

description = ''

for file_ in ('README', 'CHANGES', 'CONTRIBUTORS'):
    with open('%s.rst' % file_) as f:
        description += f.read() + '\n\n'


classifiers = ["Programming Language :: Python",
               "License :: OSI Approved :: Apache Software License",
               "Development Status :: 1 - Planning"]


setup(name='boom',
      version=__version__,
      url='https://github.com/tarekziade/boom',
      packages=find_packages(),
      long_description=description,
      description=("Simple HTTP Load tester"),
      author="Tarek Ziade",
      author_email="tarek@ziade.org",
      include_package_data=True,
      zip_safe=False,
      classifiers=classifiers,
      install_requires=install_requires,
      test_suite='unittest2.collector',
      entry_points="""
      [console_scripts]
      boom = boom._boom:main
      """)
